from typing import Any
from abc import ABC, abstractmethod
from dataclasses import dataclass
from ..network import Network
from . import label_mapping as map

@dataclass
class NodalAnalysisSolution(ABC):
    network: Network
    node_mapper: map.NetworkMapper = map.default_node_mapper
    current_source_mapper: map.SourceIndexMapper = map.alphabetic_current_source_mapper
    voltage_source_mapper: map.SourceIndexMapper = map.alphabetic_voltage_source_mapper

    @property
    def _node_mapping(self) -> map.LabelMapping:
        return self.node_mapper(self.network)

    @property
    def _current_source_mapping(self) -> map.LabelMapping:
        return self.current_source_mapper(self.network)

    @property
    def _voltage_source_mapping(self) -> map.LabelMapping:
        return self.voltage_source_mapper(self.network)

    @abstractmethod
    def get_potential(self, node_id: str) -> Any:
        ...

    @abstractmethod
    def get_current(self, branch_id: str) -> Any:
        ...

    def get_voltage(self, branch_id: str) -> Any:
        phi1 = self.get_potential(self.network[branch_id].node1)
        phi2 = self.get_potential(self.network[branch_id].node2)
        return phi1-phi2

    def get_power(self, branch_id: str) -> Any:
        return self.get_voltage(branch_id)*self.get_current(branch_id).conjugate()


  
from typing import Callable, Tuple
import numpy as np
from .state_space_model import StateSpaceModel

TimeDomainFunction = Callable[[np.ndarray], np.ndarray]
TimeDomainSeries = Tuple[np.ndarray, np.ndarray]
FrequencyDomainSeries = Tuple[np.ndarray, np.ndarray]

StateSpaceSolver = Callable[[StateSpaceModel, np.ndarray, np.ndarray, np.ndarray], tuple[np.ndarray, np.ndarray, np.ndarray]]


from ..network import Network
from ..elements import is_current_source, is_ideal_voltage_source
from typing import Callable
from dataclasses import dataclass

class DistinctValues(Exception):
    ...

@dataclass
class LabelMapping:
    mapping: dict[str, int]

    def __post_init__(self) -> None:
        if len(set(self.mapping.values())) != len(self.mapping):
            raise DistinctValues

    def __getitem__(self, label: str) -> int:
        return self.mapping[label]

    def __call__(self, *labels: str) -> tuple[int, ...]:
        return tuple(self[label] for label in labels)

    @property
    def keys(self) -> list[str]:
        return list(self.mapping.keys())

    @property
    def values(self) -> list[int]:
        return list(self.mapping.values())

    @property
    def N(self) -> int:
        return len(self.mapping)

    def __iter__(self):
        return iter(self.mapping.keys())

def filter(mapping: LabelMapping, filter_fcn: Callable[[str], bool]) -> LabelMapping:
    return LabelMapping({k: mapping[k] for k in mapping.keys if filter_fcn(k)})
    

NetworkMapper = Callable[[Network], LabelMapping]
SourceIndexMapper = Callable[[Network], LabelMapping]

def alphabetic_node_mapper(network: Network) -> LabelMapping:
    node_labels_without_zero = [label for label in sorted(network.node_labels) if label != network.node_zero_label] 
    return LabelMapping({k: v for v, k in enumerate(node_labels_without_zero)})

default_node_mapper = alphabetic_node_mapper

def alphabetic_source_mapper(network: Network) -> LabelMapping:
    current_source_labels = [b.id for b in network.branches if is_current_source(b.element)]
    voltage_source_labels = [b.id for b in network.branches if is_ideal_voltage_source(b.element)]
    sorted_soruce_labels = sorted(current_source_labels+voltage_source_labels)
    return LabelMapping({k: v for v, k in enumerate(sorted_soruce_labels)})

def alphabetic_current_source_mapper(network: Network) -> LabelMapping:
    current_source_labels = sorted([b.id for b in network.branches if is_current_source(b.element)])
    return LabelMapping({k: v for v, k in enumerate(current_source_labels)})

def alphabetic_voltage_source_mapper(network: Network) -> LabelMapping:
    voltage_source_labels = sorted([b.id for b in network.branches if is_ideal_voltage_source(b.element)])
    return LabelMapping({k: v for v, k in enumerate(voltage_source_labels)})

default_source_mapper = alphabetic_source_mapper


import numpy as np
import itertools
from dataclasses import dataclass
from .node_analysis import source_incidence_matrix, nodal_analysis_coefficient_matrix
from . import label_mapping as map
from ..elements import is_ideal_voltage_source
from ..network import Network
from ...SignalProcessing import state_space_model as sp

def state_space_matrices(network: Network, c_values: dict[str, float] = {}, l_values: dict[str, float] = {}, node_mapper: map.NetworkMapper = map.default_node_mapper, current_source_mapper: map.SourceIndexMapper = map.alphabetic_current_source_mapper, voltage_source_mapper: map.SourceIndexMapper = map.alphabetic_voltage_source_mapper) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    def element_incidence_matrix(values: dict[str, float]) -> np.ndarray:
        node_mapping = node_mapper(network)
        Delta = np.zeros((len(values), node_mapping.N))
        for (k, value), (i_label) in itertools.product(enumerate(values), node_mapping):
            if i_label == network[value].node1:
                Delta[k][node_mapping(i_label)] = +1
            if i_label == network[value].node2:
                Delta[k][node_mapping(i_label)] = -1
        return np.hstack((Delta, np.zeros((Delta.shape[0], voltage_source_mapper(network).N))))
    def source_and_inductance_incidence_matrix(values: dict[str, float]) -> tuple[np.ndarray, np.ndarray]:
        voltage_source_mapping_all = voltage_source_mapper(network)
        source_mapping_all = map.default_source_mapper(network)
        Qi = source_incidence_matrix(network=network)
        Q = np.zeros((voltage_source_mapping_all.N, voltage_source_mapping_all.N), dtype=int)
        for i in voltage_source_mapping_all.values:
            Q[i][i] = 1
        Q = np.vstack((np.hstack( (Qi, np.zeros((Qi.shape[0], Q.shape[1]) ))),
                    np.hstack( (np.zeros((Q.shape[0], Qi.shape[1])), Q) )))
        QS = Q[:,[source_mapping_all[l] for l in source_mapping_all if l not in l_values]]
        QL = Q[:,[source_mapping_all[l] for l in source_mapping_all if l in l_values]]
        return QS, QL
    def value_matrix(c_values: dict[str, float], l_values: dict[str, float]) -> np.ndarray:
        return np.vstack((
            np.hstack(( np.diag([-C for C in c_values.values()]), np.zeros((len(c_values), len(l_values))) )),
            np.hstack(( np.zeros((len(l_values), len(c_values))), np.diag([L for L in l_values.values()]) ))
        ))

    Delta = element_incidence_matrix(c_values)
    A_tilde = nodal_analysis_coefficient_matrix(network).real
    QS, QL = source_and_inductance_incidence_matrix(l_values)
    DQ = np.hstack((Delta.T, QL))
    Lambda = value_matrix(c_values, l_values)
    invLambda = np.diag([1/L for L in np.diag(Lambda)])

    inv_A_tilde = np.linalg.inv(A_tilde)
    transformed_inv_A_tilde = DQ.T @ inv_A_tilde
    sorted_A_tilde = np.linalg.inv(transformed_inv_A_tilde @ DQ)

    A = invLambda @ sorted_A_tilde
    C = transformed_inv_A_tilde.T @ sorted_A_tilde
    B = (-invLambda @ C.T) @ QS
    D = (inv_A_tilde - transformed_inv_A_tilde.T @ C.T) @ QS

    return A, B, C, D

@dataclass(frozen=True)
class NodalStateSpaceModel(sp.StateSpaceModel):
    network: Network
    c_values: dict[str, float]
    l_values: dict[str, float]
    node_index_mapping: map.LabelMapping
    voltage_source_index_mapping: map.LabelMapping
    current_source_index_mapping: map.LabelMapping

    def _row_for_potential(self, node_id: str, matrix: np.ndarray) -> np.ndarray:
        if node_id in self.node_index_mapping:
            return matrix[:][self.node_index_mapping[node_id]:self.node_index_mapping[node_id]+1]
        return np.zeros((1,matrix.shape[1]))

    def c_row_for_potential(self, node_id: str) -> np.ndarray:
        return self._row_for_potential(node_id, self.C)

    def _one_vector(_, length: int, one_indices: list[int]) -> np.ndarray:
        vector = np.zeros(length)
        vector[one_indices] = 1
        return vector

    def c_row_voltage(self, branch_id: str) -> np.ndarray:
        branch = self.network[branch_id]
        c_pos = self.c_row_for_potential(branch.node1)
        c_neg = self.c_row_for_potential(branch.node2)
        return c_pos - c_neg

    def c_row_current(self, branch_id: str) -> np.ndarray:
        voltage_source_mapping = map.filter(self.voltage_source_index_mapping, lambda x: is_ideal_voltage_source(self.network[x].element))
        if branch_id in self.c_values:
            idx = list(self.c_values.keys()).index(branch_id)
            return self.c_values[branch_id]*self.A[idx][:]
        if branch_id in self.voltage_source_index_mapping:
            return self.C[voltage_source_mapping[branch_id]+self.node_index_mapping.N][:]
        if branch_id in self.current_source_index_mapping:
            return np.zeros(self.C.shape[1])
        branch = self.network[branch_id]
        c_pos = self.c_row_for_potential(branch.node1)
        c_neg = self.c_row_for_potential(branch.node2)
        return (c_pos - c_neg)/branch.element.Z

    def d_row_for_potential(self, node_id: str) -> np.ndarray:
        return self._row_for_potential(node_id, self.D)

    def d_row_voltage(self, branch_id: str) -> np.ndarray:
        branch = self.network[branch_id]
        d_pos = self.d_row_for_potential(branch.node1)
        d_neg = self.d_row_for_potential(branch.node2)
        return d_pos - d_neg

    def d_row_current(self, branch_id: str) -> np.ndarray:
        if branch_id in self.c_values:
            idx = list(self.c_values.keys()).index(branch_id)
            return self.c_values[branch_id]*self.B[idx][:]
        if branch_id in self.voltage_source_index_mapping:
            return self.D[self.voltage_source_index_mapping[branch_id]+self.node_index_mapping.N][:]
        if branch_id in self.current_source_index_mapping:
            d_row = np.zeros(self.D.shape[1])
            d_row[self.current_source_index_mapping[branch_id]] = 1
            return d_row
        branch = self.network[branch_id]
        d_pos = self.d_row_for_potential(branch.node1)
        d_neg = self.d_row_for_potential(branch.node2)
        return (d_pos - d_neg)/branch.element.Z

    @property
    def sources(self) -> list[str]:
        current_sources = self.current_source_index_mapping.keys
        voltage_sources = [vs for vs in self.voltage_source_index_mapping.keys if vs not in self.l_values]
        return current_sources+voltage_sources

def nodal_state_space_model(network: Network, c_values: dict[str, float] = {}, l_values: dict[str, float] = {}, node_index_mapper: map.NetworkMapper = map.default_node_mapper, voltage_source_index_mapper: map.SourceIndexMapper = map.alphabetic_voltage_source_mapper, current_source_index_mapper: map.SourceIndexMapper = map.alphabetic_current_source_mapper) -> NodalStateSpaceModel:
    A, B, C, D = state_space_matrices(
        network=network,
        c_values=c_values,
        l_values=l_values,
        node_mapper=node_index_mapper,
        voltage_source_mapper=voltage_source_index_mapper,
        current_source_mapper=current_source_index_mapper
    )
    return NodalStateSpaceModel(
        A=A, B=B, C=C, D=D,
        network=network,
        c_values=c_values,
        l_values=l_values,
        node_index_mapping=node_index_mapper(network),
        voltage_source_index_mapping=voltage_source_index_mapper(network),
        current_source_index_mapping=current_source_index_mapper(network),
    )


from .circuit import Circuit, transform_circuit
import numpy as np
from ..Network.NodalAnalysis.state_space_model import nodal_state_space_model
from ..SignalProcessing.state_space_model import StateSpaceModel

def state_space_model(circuit: Circuit, potential_nodes: list[str] = [], voltage_ids: list[str] = [], current_ids: list[str] = []) -> StateSpaceModel:
    ssm = nodal_state_space_model(
        network=transform_circuit(circuit, w=0),
        c_values={C.id : float(C.value['C']) for C in [c for c in circuit.components if c.type == 'capacitor']},
        l_values={L.id : float(L.value['L']) for L in [c for c in circuit.components if c.type == 'inductance']}
    )

    C = np.ndarray(shape=(0, ssm.n_states))
    for id in potential_nodes:
        C = np.vstack([C, ssm.c_row_for_potential(id)])
    for id in voltage_ids:
        C = np.vstack([C, ssm.c_row_voltage(id)])
    for id in current_ids:
        C = np.vstack([C, ssm.c_row_current(id)])

    D = np.ndarray(shape=(0, ssm.n_inputs))
    for id in potential_nodes:
        D = np.vstack([D, ssm.d_row_for_potential(id)])
    for id in voltage_ids:
        D = np.vstack([D, ssm.d_row_voltage(id)])
    for id in current_ids:
        D = np.vstack([D, ssm.d_row_current(id)])

    return StateSpaceModel(
        A=ssm.A,
        B=ssm.B,
        C=C,
        D=D
    )


import numpy as np
from ..Network import elements as elm
from ..Network import network as ntw
from typing import Callable, TypeVar
from . import components as ccp
from ..SignalProcessing.periodic_functions import periodic_function, fourier_series

CircuitComponent = TypeVar("CircuitComponent", bound=ccp.Component)
CircuitComponentTranslator = Callable[[ccp.Component, float, float], ntw.Branch]

def resistor(resistor: ccp.Component, *_) -> ntw.Branch:
    R = float(resistor.value['R'])
    return ntw.Branch(resistor.nodes[0], resistor.nodes[1], elm.resistor(resistor.id, R))

def impedance(impedance: ccp.Component, *_) -> ntw.Branch:
    Z = complex(
        float(impedance.value['R']),
        float(impedance.value['X'])
    )
    return ntw.Branch(impedance.nodes[0], impedance.nodes[1], elm.impedance(impedance.id, Z))

def capacitor(capacitor: ccp.Component, w: float = 0, *_) -> ntw.Branch:
    C = float(capacitor.value['C'])
    return ntw.Branch(
        capacitor.nodes[0],
        capacitor.nodes[1],
        elm.admittance(capacitor.id, elm.admittance_value(B=w*C)))

def inductance(inductance: ccp.Component, w: float = 0, *_) -> ntw.Branch:
    L = float(inductance.value['L'])
    return ntw.Branch(
        inductance.nodes[0],
        inductance.nodes[1],
        elm.impedance(inductance.id, elm.impedance_value(X=w*L))
    )

def dc_voltage_source(voltage_source: ccp.Component, w: float = 0, w_resolution: float = 1e-3) -> ntw.Branch:
    V = elm.complex_value(float(voltage_source.value['V']), 0)
    Z = elm.complex_value(float(voltage_source.value['R']), 0)
    element = elm.voltage_source(voltage_source.id, V, Z)
    if np.abs(w-float(voltage_source.value['w'])) > w_resolution:
        element = elm.short_circuit(voltage_source.id)
    return ntw.Branch(
        voltage_source.nodes[0],
        voltage_source.nodes[1],
        element)

def ac_voltage_source(voltage_source: ccp.Component, w: float = 0, w_resolution: float = 1e-3) -> ntw.Branch:
    vs_V = float(voltage_source.value['V'])
    vs_phi = float(voltage_source.value['phi'])
    vs_R = float(voltage_source.value['R'])
    element = elm.voltage_source(voltage_source.id, elm.complex_value(vs_V, vs_phi), elm.complex_value(vs_R, 0))
    if np.abs(w-float(voltage_source.value['w'])) > w_resolution:
        element = elm.short_circuit(voltage_source.id)
    return ntw.Branch(
        voltage_source.nodes[0],
        voltage_source.nodes[1],
        element)

def complex_voltage_source(voltage_source: ccp.Component, *_) -> ntw.Branch:
    cs_V = complex(
        float(voltage_source.value['V_real']),
        float(voltage_source.value['V_imag'])
    )
    cs_Z = complex(
        float(voltage_source.value['R']),
        float(voltage_source.value['X'])
    )
    element = elm.voltage_source(voltage_source.id, cs_V, cs_Z)
    return ntw.Branch(
        voltage_source.nodes[0],
        voltage_source.nodes[1],
        element)

def periodic_voltage_source(source: ccp.Component, w: float = 0, w_resolution: float = 1e-3) -> ntw.Branch:
    wavetype = str(source.value['wavetype'])
    w0 = float(source.value['w'])
    V = float(source.value['V'])
    phi = float(source.value['phi'])
    frequency_properties = fourier_series(periodic_function(wavetype)(period=2*np.pi/w0, amplitude=V, phase=phi))
    n = np.round(w/w0)
    delta_n = np.abs(w/w0 - n)
    if delta_n > w_resolution/w0:
        return ntw.Branch(
            source.nodes[0],
            source.nodes[1],
            elm.short_circuit(source.id))
    single_frequency_source = ccp.ac_voltage_source(
        id=source.id,
        nodes=(source.nodes[0], source.nodes[1]),
        w=w,
        phi=frequency_properties.phase(n),
        V=frequency_properties.amplitude(n)
    )
    return ac_voltage_source(single_frequency_source, w, w_resolution)

def dc_current_source(current_source: ccp.Component, w: float = 0, w_resolution: float = 1e-3) -> ntw.Branch:
    cs_I = float(current_source.value['I'])
    cs_G = float(current_source.value['G'])
    cs_w = float(current_source.value['w'])
    element = elm.current_source(current_source.id, elm.complex_value(cs_I, 0), elm.complex_value(cs_G, 0))
    if np.abs(w-cs_w) > w_resolution:
        element = elm.open_circuit(current_source.id)
    return ntw.Branch(
        current_source.nodes[0],
        current_source.nodes[1],
        element
    )

def ac_current_source(current_source: ccp.Component, w: float = 0, w_resolution: float = 1e-3) -> ntw.Branch:
    cs_I = float(current_source.value['I'])
    cs_G = float(current_source.value['G'])
    cs_w = float(current_source.value['w'])
    cs_phi = float(current_source.value['phi'])
    element = elm.current_source(current_source.id, elm.complex_value(cs_I, cs_phi), elm.complex_value(cs_G))
    if np.abs(w-cs_w) > w_resolution:
        element = elm.open_circuit(current_source.id)
    return ntw.Branch(
        current_source.nodes[0],
        current_source.nodes[1],
        element
    )

def complex_current_source(current_source: ccp.Component, w: float = 0, w_resolution: float = 1e-3) -> ntw.Branch:
    cs_I = complex(
        float(current_source.value['I_real']),
        float(current_source.value['I_imag'])
    )
    cs_Y = complex(
        float(current_source.value['G']),
        float(current_source.value['B'])
    )
    cs_w = float(current_source.value['w'])
    element = elm.current_source(current_source.id, cs_I, cs_Y)
    if np.abs(w-cs_w) > w_resolution:
        element = elm.short_circuit(current_source.id)
    return ntw.Branch(
        current_source.nodes[0],
        current_source.nodes[1],
        element)

def periodic_current_source(source: ccp.Component, w: float = 0, w_resolution: float = 1e-3) -> ntw.Branch:
    wavetype = str(source.value['wavetype'])
    w0 = float(source.value['w'])
    I = float(source.value['I'])
    phi = float(source.value['phi'])
    frequency_properties = fourier_series(periodic_function(wavetype)(period=2*np.pi/w0, amplitude=I, phase=phi))
    n = np.round(w/w0)
    delta_n = np.abs(w/w0 - n)
    if delta_n > w_resolution/w0:
        return ntw.Branch(
            source.nodes[0],
            source.nodes[1],
            elm.open_circuit(source.id))
    single_frequency_source = ccp.ac_current_source(
        id=source.id,
        nodes=(source.nodes[0], source.nodes[1]),
        w=w,
        phi=frequency_properties.phase(n),
        I=frequency_properties.amplitude(n)
    )
    return ac_current_source(single_frequency_source, w, w_resolution)

def short_circuit(short_circuit: ccp.Component, *_) -> ntw.Branch:
    return ntw.Branch(
        short_circuit.nodes[0],
        short_circuit.nodes[1],
        elm.short_circuit(short_circuit.id)
    )

def resistive_load(load: ccp.Component, *_) -> ntw.Branch:
    return ntw.Branch(
        load.nodes[0],
        load.nodes[1],
        elm.load(load.id, float(load.value['P']), float(load.value['V_ref']))
    )

transformers : dict[str, CircuitComponentTranslator] = {
    'resistor' : resistor,
    'impedance' : impedance,
    'capacitor' : capacitor,
    'inductance' : inductance,
    'dc_voltage_source' : dc_voltage_source,
    'ac_voltage_source' : ac_voltage_source,
    'complex_voltage_source' : complex_voltage_source,
    'dc_current_source' : dc_current_source,
    'ac_current_source' : ac_current_source,
    'complex_current_source' : complex_current_source,
    'periodic_voltage_source' : periodic_voltage_source,
    'periodic_current_source' : periodic_current_source,
    'short_circuit' : short_circuit,
    'resistive_load' : resistive_load,
    'lamp' : resistive_load
}


from .circuit import Circuit, transform, frequency_components
from ..SignalProcessing.types import TimeDomainFunction, FrequencyDomainSeries, TimeDomainSeries, StateSpaceSolver
from ..SignalProcessing.state_space_model import StateSpaceModel, continuous_state_space_solver
from ..Network.NodalAnalysis.bias_point_analysis import nodal_analysis_bias_point_solver
from ..Network.NodalAnalysis.state_space_model import nodal_state_space_model
from ..Network.solution import NetworkSolver
from typing import Any
from dataclasses import dataclass, field
from abc import ABC, abstractmethod
import numpy as np

@dataclass
class CircuitSolution(ABC):
    circuit: Circuit

    @abstractmethod
    def get_voltage(self, id: str) -> Any:
        ...

    @abstractmethod
    def get_current(self, id: str) -> Any:
        ...

    @abstractmethod
    def get_potential(self, id: str) -> Any:
        ...

    @abstractmethod
    def get_power(self, id: str) -> Any:
        ...

@dataclass
class DCSolution(CircuitSolution):
    solver: NetworkSolver = field(default=nodal_analysis_bias_point_solver)

    def __post_init__(self):
        network = transform(self.circuit, w=[0])[0]
        self._solution = self.solver(network)

    def get_voltage(self, component_id: str) -> float:
        return self._solution.get_voltage(component_id).real

    def get_current(self, component_id: str) -> float:
        return self._solution.get_current(component_id).real

    def get_potential(self, node_id: str) -> float:
        return self._solution.get_potential(node_id).real

    def get_power(self, component_id: str) -> float:
        return self.get_voltage(component_id)*self.get_current(component_id)

@dataclass
class ComplexSolution(CircuitSolution):
    solver: NetworkSolver = field(default=nodal_analysis_bias_point_solver)
    w: float = 0
    peak_values: bool = False

    def __post_init__(self):
        network = transform(self.circuit, w=[self.w])[0]
        self._solution = self.solver(network)

    def get_voltage(self, component_id: str) -> complex:
        if self.peak_values:
            return self._solution.get_voltage(component_id)
        return self._solution.get_voltage(component_id)/np.sqrt(2)

    def get_current(self, component_id: str) -> complex:
        if self.peak_values:
            return self._solution.get_current(component_id)
        return self._solution.get_current(component_id)/np.sqrt(2)

    def get_potential(self, node_id: str) -> complex:
        if self.peak_values:
            return self._solution.get_potential(node_id)
        return self._solution.get_potential(node_id)/np.sqrt(2)

    def get_power(self, component_id: str) -> complex:
        if self.peak_values:
            return 1/2*self.get_voltage(component_id)*np.conj(self.get_current(component_id))
        return self.get_voltage(component_id)*np.conj(self.get_current(component_id))

@dataclass
class TimeDomainSolution(CircuitSolution):
    w_max: float = field(default=0)
    w: list[float] = field(default_factory=list, init=False)
    solver: NetworkSolver = field(default=nodal_analysis_bias_point_solver)

    def __post_init__(self):
        self.w = frequency_components(self.circuit, self.w_max)
        networks = transform(self.circuit, w=self.w)
        self._solutions = [self.solver(network) for network in networks]

    def get_voltage(self, component_id: str) -> TimeDomainFunction:
        voltages = [solution.get_voltage(component_id) for solution in self._solutions]
        return np.vectorize(lambda t: np.array(np.sum([np.abs(V)*np.cos(w*t+np.angle(V)) for V, w in zip(voltages, self.w)])))

    def get_current(self, component_id: str) -> TimeDomainFunction:
        currents = [solution.get_current(component_id) for solution in self._solutions]
        return np.vectorize(lambda t: np.array(np.sum([np.abs(V)*np.cos(w*t+np.angle(V)) for V, w in zip(currents, self.w)])))

    def get_potential(self, node_id: str) -> TimeDomainFunction:
        potentials = [solution.get_potential(node_id) for solution in self._solutions]
        return np.vectorize(lambda t: np.array(np.sum([np.abs(phi)*np.cos(w*t+np.angle(phi)) for phi, w in zip(potentials, self.w)])))

    def get_power(self, component_id: str) -> TimeDomainFunction:
        voltage = self.get_voltage(component_id)
        current = self.get_current(component_id)
        return lambda t: np.array(voltage(t))*np.array(current(t))

@dataclass
class FrequencyDomainSolution(CircuitSolution):
    w_max: float = field(default=0)
    solver: NetworkSolver = field(default=nodal_analysis_bias_point_solver)
    one_sided: bool = field(default=True)

    def __post_init__(self):
        self.w = np.array(frequency_components(self.circuit, self.w_max))
        self._solutions = np.array([ComplexSolution(circuit=self.circuit, solver=self.solver, w=w, peak_values=True) for w in self.w])
        if not self.one_sided:
            self.w = np.concatenate((-self.w[-1:0:-1], self.w))
            self._solutions = 1/2*np.concatenate((np.conj(self._solutions[-1:0:-1]), self._solutions))

    def get_voltage(self, component_id: str) -> FrequencyDomainSeries:
        voltages = np.array([solution.get_voltage(component_id) for solution in self._solutions])
        return np.array(self.w), voltages

    def get_current(self, component_id: str) -> FrequencyDomainSeries:
        currents = np.array([solution.get_current(component_id) for solution in self._solutions])
        return np.array(self.w), currents

    def get_potential(self, node_id: str) -> FrequencyDomainSeries:
        potentials = np.array([solution.get_potential(node_id) for solution in self._solutions])
        return np.array(self.w), potentials

    def get_power(self, component_id: str) -> FrequencyDomainSeries:
        power = np.array([solution.get_power(component_id) for solution in self._solutions])
        return np.array(self.w), power

@dataclass
class TransientSolution(CircuitSolution):
    tin: np.ndarray = field(default_factory=lambda : np.empty((0,1)))
    input: dict[str, TimeDomainFunction] = field(default_factory=dict)
    solver: StateSpaceSolver = field(default=continuous_state_space_solver)

    def __post_init__(self):
        network = transform(self.circuit, w=[0])[0]

        C_values = {c.id: float(c.value['C']) for c in self.circuit.components if c.type == 'capacitor'}
        L_values = {c.id: float(c.value['L']) for c in self.circuit.components if c.type == 'inductance'}

        self._ssm = nodal_state_space_model(network, c_values=C_values, l_values=L_values)
        self._u = np.array([self.input[input_id](self.tin) for input_id in self._ssm.sources])
        self._tout, self._x, _ = self.solver(
            StateSpaceModel(A=self._ssm.A, B=self._ssm.B, C=np.eye(self._ssm.A.shape[0]), D=np.zeros((self._ssm.A.shape[0], self._ssm.B.shape[1]))),
            self._u.T,
            self.tin,
            np.zeros((self._ssm.A.shape[0], 1))
        )
        self._x = np.reshape(self._x, (self._x.shape[0], self._ssm.A.shape[0])).T

    @property
    def t(self) -> np.ndarray:
        return self._tout

    def get_potential(self, node_id: str) -> TimeDomainSeries:
        return self._tout, np.reshape(self._ssm.c_row_for_potential(node_id)@self._x + self._ssm.d_row_for_potential(node_id)@self._u, (-1,))

    def get_voltage(self, component_id: str) -> TimeDomainSeries:
        return self._tout, np.reshape(self._ssm.c_row_voltage(component_id)@self._x + self._ssm.d_row_voltage(component_id)@self._u, (-1,))

    def get_current(self, component_id: str) -> TimeDomainSeries:
        return self._tout, np.reshape(self._ssm.c_row_current(component_id)@self._x + self._ssm.d_row_current(component_id)@self._u, (-1,))

    def get_power(self, component_id: str) -> TimeDomainSeries:
        return self._tout, self.get_voltage(component_id)[1]*self.get_current(component_id)[1]
        
        


from typing import Any, Callable
from .circuit import Circuit, Component
from . import components as ccp
from dataclasses import asdict
import json

class UnidentifiedComponent(Exception):
    ...

class UnknownCircuitComponent(Exception):
    ...

class IncorrectComponentInformation(Exception):
    ...

circuit_component_translators : dict[str, Callable[..., ccp.Component]] = {
    "resistor" : ccp.resistor,
    "conductance" : ccp.conductance,
    "impedance" : ccp.impedance,
    "admittance" : ccp.admittance,
    "dc_voltage_source" : ccp.dc_voltage_source,
    "ac_voltage_source" : ccp.ac_voltage_source,
    "complex_voltage_source" : ccp.complex_voltage_source,
    "dc_current_source" : ccp.dc_current_source,
    "ac_current_source" : ccp.ac_current_source,
    "complex_current_source" : ccp.complex_current_source
}

def generate_component(component: dict[str, Any]) -> Component:
    component = component.copy()
    try:
        component_id = component['id']
    except KeyError:
        raise UnidentifiedComponent(f'Unidentified component')
    try:
        component_value = component.pop('value')
    except KeyError:
        raise IncorrectComponentInformation(f"Missing value of component '{component_id}'.")
    try:
        component_type = component.pop('type')
    except KeyError:
        raise IncorrectComponentInformation(f"Missing type information of component '{component_id}'.")
    try:
        component_nodes = component['nodes']
    except KeyError:
        raise IncorrectComponentInformation(f"Missing node information of component '{component_id}'.")
    try:
        component_factory = circuit_component_translators[component_type]
    except KeyError:
        raise UnknownCircuitComponent(f"Unknown type '{component_type}' of component '{component_id}' is unknown.")
    try:
        return component_factory(id=component_id, nodes=component_nodes, **component_value)
    except TypeError:
        raise IncorrectComponentInformation(f"Given value information for component '{component_id}' of type '{component_type}' is incorrect: '{component_value}'")

def load_circuit(circuit_dict: list[dict[str, Any]]) -> Circuit:
    return Circuit([generate_component(entry) for entry in circuit_dict])

def dump_circuit(circuit: Circuit) -> list[dict[str, Any]]:
    return [asdict(c) for c in circuit.components]

def load_circuit_from_json(filename: str) -> Circuit:
    with open(filename, 'r') as json_file:
        circuit_dict = json.load(json_file)
    return load_circuit(circuit_dict)

def dump_circuit_to_json(circuit: Circuit, filename: str) -> None:
    with open(filename, 'w') as json_file:
        json.dump(dump_circuit(circuit), json_file)


import numpy as np
from dataclasses import dataclass
import scipy.signal

@dataclass(frozen=True)
class StateSpaceModel:
    A: np.ndarray
    B: np.ndarray
    C: np.ndarray
    D: np.ndarray

    def __post_init__(self) -> None:
        if self.A.shape[0] != self.A.shape[1]:
            raise ValueError('Matrix A must be square')
        if self.B.shape[0] != self.A.shape[0]:
            raise ValueError('Number of columns of matrix B must be equal to number of rows of matrix A')
        if self.C.shape[1] != self.A.shape[0]:
            raise ValueError('Number of columns of matrix B must be equal to number of rows of matrix A')
        if self.D.shape[0] != self.C.shape[0]:
            raise ValueError('Number of rows of matrix D must be equal to number of rows of matrix C')
        if self.D.shape[1] != self.B.shape[1]:
            raise ValueError('Number of columns of matrix D must be equal to number of columns of matrix B')

    @property
    def n_states(self) -> int:
        return self.A.shape[0]

    @property
    def n_inputs(self) -> int:
        return self.B.shape[1]

    @property
    def n_outputs(self) -> int:
        return self.C.shape[0]

def continuous_state_space_solver(ssm: StateSpaceModel, y: np.ndarray, t: np.ndarray, x0: np.ndarray) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    sys = scipy.signal.StateSpace(ssm.A, ssm.B, ssm.C, ssm.D)
    return scipy.signal.lsim(sys, y, t)


from . import elements as elm
from dataclasses import dataclass

class FloatingGroundNode(Exception): pass

class AmbiguousBranchIDs(Exception): pass

@dataclass(frozen=True)
class Branch:
    node1 : str
    node2 : str
    element : elm.NortenTheveninElement

    @property
    def id(self) -> str:
        return self.element.name

@dataclass(frozen=True)
class Network:
    branches: list[Branch]
    node_zero_label: str = '0'

    def __post_init__(self):
        if self.node_zero_label not in self.node_labels and self.number_of_nodes != 0:
            raise FloatingGroundNode
        if len(set(self.branch_ids)) != len(self.branches):
            raise AmbiguousBranchIDs
        
    @property
    def branch_ids(self) -> list[str]:
        return [b.id for b in self.branches]

    @property
    def node_labels(self) -> list[str]:
        if len(self.branches) == 0:
            return [self.node_zero_label]
        node1_set = {branch.node1 for branch in self.branches}
        node2_set = {branch.node2 for branch in self.branches}
        return sorted(list(node1_set.union(node2_set)))

    @property
    def number_of_nodes(self) -> int:
        return len(self.node_labels)

    def is_zero_node(self, node: str) -> bool:
        return node == self.node_zero_label

    def branches_connected_to(self, node: str) -> list[Branch]:
        connected_branches = [branch for branch in self.branches if branch.node1 == node or branch.node2 == node]
        connected_branches.sort(key=lambda x: x.node1 if x.node1!=node else x.node2)
        return connected_branches

    def nodes_connected_to(self, node: str) -> set[str]:
        return {b.node1 if b.node1 != node else b.node2 for b in self.branches_connected_to(node=node)}

    def branches_between(self, node1: str, node2: str) -> list[Branch]:
        return [branch for branch in self.branches if set((branch.node1, branch.node2)) == set((node1, node2))]

    def __getitem__(self, id: str) -> Branch:
        return {b.id: b for b in self.branches}[id]

def ideal_voltage_sources(network: Network) -> list[Branch]:
    return [b for b in network.branches if elm.is_ideal_voltage_source(b.element)]

def ideal_current_sources(network: Network) -> list[Branch]:
    return [b for b in network.branches if elm.is_ideal_current_source(b.element)]

def passive_elements(network: Network) -> list[Branch]:
    return [b for b in network.branches if not elm.is_active(b.element)]


import numpy as np
from dataclasses import dataclass, field

@dataclass(frozen=True)
class FloatPrecision:
    value: float
    precision: int = 3
    min_exp: int = -16
    max_exp: int = 16

    @property
    def is_zero(self) -> bool:
        if self.value == 0:
            return True
        return self.exponent < self.min_exp

    @property
    def is_inf(self) -> bool:
        return self.exponent > self.max_exp

    @property
    def mantissa(self) -> int:
        return int(np.round(self.value/(10**self.exponent)))

    @property
    def exponent(self) -> int:
        abs_value = abs(self.value)
        str_abs_value = self._float_to_string(abs_value)
        pre_decimal = str_abs_value.split('.')[0]
        if pre_decimal == '0':
            post_decimal = str_abs_value.split('.')[-1]
            exp0 = len(post_decimal)-len(post_decimal.lstrip('0'))
            rounded_value = np.round(abs_value * 10**exp0, decimals=self.precision) / 10**exp0
            rounded_post_decimal = self._float_to_string(rounded_value).split('.')[-1]
            if rounded_post_decimal == '0':
                return 0
            return -(len(rounded_post_decimal)-len(rounded_post_decimal.lstrip('0'))+self.precision)
        return len(pre_decimal)-self.precision

    def _float_to_string(self, value: float) -> str:
        str_abs_value = str(value)
        if str_abs_value.count('e') == 1:
            post_decimal_digits = abs(int(str_abs_value.split('e')[-1]))+1+self.precision
            return f'{value:.{post_decimal_digits}f}'
        return str(value)

class Float3(FloatPrecision):
    @property
    def mantissa3(self) -> float:
        return self.mantissa * 10**(self.exponent-self.exponent3)

    @property
    def exponent3(self) -> int:
        return int(3*np.floor((self.precision + self.exponent - 1)/3))

@dataclass
class ScientificFloat:
    value: float
    unit: str = ''
    precision: int = 3
    use_exp_prefix: bool = False
    exp_prefixes: dict[int, str] = field(default_factory=lambda: {
        -12 : 'p',
        -9 : 'n',
        -6 : 'u',
        -3 : 'm',
        -1 : 'c',
        3 : 'k',
        6 : 'M',
        9 : 'G',
        12 : 'T'
    })

    @property
    def value3(self) -> Float3:
        if self.use_exp_prefix:
            return Float3(value=self.value, precision=self.precision, min_exp=min(self.exp_prefixes.keys()), max_exp=max(self.exp_prefixes.keys()))
        return Float3(value=self.value, precision=self.precision)

    def exp_prefix(self, exp) -> str:
        if not self.use_exp_prefix:
            return ''
        if exp > max(self.exp_prefixes.keys()):
            return self.exp_prefixes[max(self.exp_prefixes.keys())]
        if exp < min(self.exp_prefixes.keys()):
            return self.exp_prefixes[min(self.exp_prefixes.keys())]
        if exp in self.exp_prefixes.keys():
            return self.exp_prefixes.get(exp, '')
        return ''

    def exp_extension(self, exp) -> str:
        def rebase_exp(exp: int) -> int:
            if not self.use_exp_prefix:
                return exp
            if not exp in self.exp_prefixes.keys():
                if exp > max(self.exp_prefixes.keys()):
                    return exp - max(self.exp_prefixes.keys())
                if exp < min(self.exp_prefixes.keys()):
                    return exp - min(self.exp_prefixes.keys())
            return 0
        if rebase_exp(exp) == 0:
            return ''
        return f'e{rebase_exp(exp)}'

    def __str__(self) -> str:
        if self.value3.is_inf:
            return '∞' if self.value3.mantissa >= 0 else '-∞'
        pre_decimal_positions = 0 if np.abs(self.value3.mantissa3) < 1 else len(str(abs(self.value3.mantissa3)).split('.')[0])
        post_decimal_positions = max(self.precision - pre_decimal_positions, 0)
        pre_decimal = int(self.value3.mantissa3)
        pre_decimal_str = f'{pre_decimal:d}'
        post_decimal = int(np.round(np.abs(self.value3.mantissa3) % 1 * 10**(post_decimal_positions)))
        post_decimal_str = '' if post_decimal_positions == 0 else f'.{post_decimal:0{post_decimal_positions}d}'
        return f'{pre_decimal_str}{post_decimal_str}{self.exp_extension(self.value3.exponent3)}{self.exp_prefix(self.value3.exponent3)}{self.unit}'

@dataclass(frozen=True)
class ScientificComplex:
    value: complex
    unit: str = ''
    precision: int = 3
    use_exp_prefix: bool = False
    compact: bool = False
    polar: bool = False
    deg: bool = False
    exp_prefixes: dict[int, str] = field(default_factory=lambda: {
        -12 : 'p',
        -9 : 'n',
        -6 : 'u',
        -3 : 'm',
        -1 : 'c',
        3 : 'k',
        6 : 'M',
        9 : 'G',
        12 : 'T'
    })

    @property
    def real(self) -> ScientificFloat:
        return ScientificFloat(abs(self.value.real), self.unit, self.precision, self.use_exp_prefix, self.exp_prefixes)

    @property
    def imag(self) -> ScientificFloat:
        return ScientificFloat(abs(self.value.imag), self.unit, self.precision, self.use_exp_prefix, self.exp_prefixes)

    @property
    def abs(self) -> ScientificFloat:
        return ScientificFloat(abs(self.value), self.unit, self.precision, self.use_exp_prefix, self.exp_prefixes)

    @property
    def angle(self) -> float:
        return float(np.angle(self.value, deg=self.deg))
    
    @property
    def real_sign(self) -> str:
        sign = '' if self.value.real >= 0 else '- '
        if self.compact:
            return sign.strip()
        return sign
    
    @property
    def imag_sign(self) -> str:
        sign = ' + ' if self.value.imag >= 0 else ' - '
        if self.compact:
            return sign.strip()
        return sign

    def __str__(self) -> str:
        with np.errstate(divide='ignore'):
            if self.polar:
                if self.deg:
                    if np.log10(np.abs(self.angle)) <= -2:
                        return f'{str(self.abs)}'
                    return f'{str(self.abs)}∠{self.angle:.2f}'
                if np.log10(np.abs(self.angle)) <= -5:
                    return f'{str(self.abs)}'
                return f'{str(self.abs)}∠{self.angle:.4f}'
            if self.imag.value3.is_zero:
                return f'{self.real_sign}{self.real.__str__()}'
            if self.real.value3.is_zero:
                return f'{self.imag_sign}j{self.imag.__str__()}' if self.value.imag < 0 else f'j{self.imag.__str__()}'
            return f'{self.real_sign}{self.real.__str__()}{self.imag_sign}j{self.imag.__str__()}'



import numpy as np

from ..network import Network
from ..elements import is_ideal_voltage_source, is_current_source
from . import label_mapping as map
from .. import transformers as trf
import itertools

class DimensionError(Exception):
    ...

def admittance_connected_to(network: Network, node: str) -> complex:
    return sum(b.element.Y for b in network.branches_connected_to(node) if np.isfinite(b.element.Y))

def admittance_between(network: Network, node1: str, node2: str) -> complex:
    return sum([b.element.Y for b in network.branches_between(node1, node2) if np.isfinite(b.element.Y)])

def connected_nodes(network: Network, node: str) -> list[str]:
    return [b.node1 if b.node1 != node else b.node2 for b in network.branches_connected_to(node)]

def node_admittance_matrix(network: Network, node_index_mapper: map.NetworkMapper = map.default_node_mapper) -> np.ndarray:
    def node_matrix_element(i_label:str, j_label:str) -> complex:
        if i_label == j_label:
            return admittance_connected_to(no_voltage_sources_network, i_label)
        return -admittance_between(no_voltage_sources_network, i_label, j_label)
    node_mapping = node_index_mapper(network)
    no_voltage_sources_network = Network(branches=[b for b in network.branches if not is_ideal_voltage_source(b.element)], node_zero_label=network.node_zero_label)
    Y = np.zeros((node_mapping.N, node_mapping.N), dtype=complex)
    for i_label, j_label in itertools.product(node_mapping, repeat=2):
        Y[node_mapping(i_label, j_label)] = node_matrix_element(i_label, j_label)
    return Y

def voltage_source_incidence_matrix(network: Network, node_mapper: map.NetworkMapper = map.default_node_mapper, source_mapper: map.SourceIndexMapper = map.alphabetic_voltage_source_mapper) -> np.ndarray:
    def voltage_source_direction(voltage_source: str, node: str) -> int:
        if network[voltage_source].node1 == node:
            return 1
        if network[voltage_source].node2 == node:
            return -1
        return 0
    node_index = node_mapper(network)
    vs_index = source_mapper(network)
    A = np.zeros((node_index.N, vs_index.N))
    for node, vs in itertools.product(node_index.keys, vs_index.keys):
        A[node_index[node], vs_index[vs]] = voltage_source_direction(vs, node)
    return A

def nodal_analysis_coefficient_matrix(network: Network, node_mapper: map.NetworkMapper = map.default_node_mapper, source_mapper: map.SourceIndexMapper = map.alphabetic_voltage_source_mapper) -> np.ndarray:
    Y = node_admittance_matrix(network, node_mapper)
    B = voltage_source_incidence_matrix(network, node_mapper, source_mapper)
    Z = np.zeros((B.shape[1], B.shape[1]))
    return np.vstack((np.hstack((Y, B)), np.hstack((B.T, Z))))

def source_incidence_matrix(network: Network, node_mapper: map.NetworkMapper = map.default_node_mapper, source_mapper: map.SourceIndexMapper = map.alphabetic_current_source_mapper) -> np.ndarray:
    node_index = node_mapper(network)
    cs_index = source_mapper(network)
    Q = np.zeros((node_index.N, cs_index.N))
    for cs in cs_index.keys:
        source_element = network[cs]
        if network.node_zero_label != source_element.node1:
            Q[node_index[source_element.node1]][cs_index[cs]] = -1
        if network.node_zero_label != network[cs].node2:
            Q[node_index[source_element.node2]][cs_index[cs]] = 1
    return Q

def current_source_vector(network: Network, source_mapper: map.SourceIndexMapper = map.alphabetic_current_source_mapper) -> np.ndarray:
    cs_index = map.filter(source_mapper(network), lambda x: is_current_source(network[x].element))
    return np.array([network[x].element.I for x in cs_index.keys])

def current_source_incidence_vector(network: Network, node_mapper: map.NetworkMapper = map.default_node_mapper, source_mapper: map.SourceIndexMapper = map.alphabetic_current_source_mapper) -> np.ndarray:
    Q = source_incidence_matrix(network, node_mapper=node_mapper, source_mapper=source_mapper)
    Is = current_source_vector(network, source_mapper=source_mapper)
    return Q@Is

def nodal_analysis_constants_vector(network: Network, node_mapper: map.NetworkMapper = map.default_node_mapper, current_source_mapper: map.SourceIndexMapper = map.alphabetic_current_source_mapper, voltage_source_mapper: map.SourceIndexMapper = map.alphabetic_voltage_source_mapper) -> np.ndarray:
    I = current_source_incidence_vector(network, node_mapper, current_source_mapper)

    vs_mapping = voltage_source_mapper(network)
    V = np.array([network[vs].element.V for vs in vs_mapping.keys])
    return np.hstack((I, V))

def open_circuit_impedance(network: Network, node1: str, node2: str, node_index_mapper: map.NetworkMapper = map.default_node_mapper) -> complex:
    if node1 == node2:
        return 0
    if any([is_ideal_voltage_source(b.element) for b in network.branches_between(node1, node2)]):
        return 0
    if network.is_zero_node(node1):
        node1, node2 = node2, node1
    network = trf.switch_ground_node(network=network, new_ground=node2)
    Y = node_admittance_matrix(network, node_index_mapper=node_index_mapper)
    Y = np.delete(Y, np.where(~Y.any(axis=0))[0], axis=1)
    Y = np.delete(Y, np.where(~Y.any(axis=1))[0], axis=0)
    Z = np.linalg.inv(Y)
    i1 = node_index_mapper(network)[node1]
    return Z[i1][i1]

def element_impedance(network: Network, element: str, node_index_mapper: map.NetworkMapper = map.default_node_mapper) -> complex:
    return open_circuit_impedance(
        network=trf.remove_element(network, element),
        node1=network[element].node1,
        node2=network[element].node2,
        node_index_mapper=node_index_mapper
    )

def calculate_node_voltages0(Y : np.ndarray, I : np.ndarray) -> np.ndarray:
    if np.any(np.logical_not(np.isfinite(Y))):
        raise ValueError
    if np.any(np.logical_not(np.isfinite(I))):
        raise ValueError
    if Y.ndim != 2:
        raise DimensionError('dim error')
    if I.ndim != 1:
        raise DimensionError('dim error')
    m, n = Y.shape
    if n != m:
        raise DimensionError('dim error')
    return np.linalg.solve(Y, I)



from .Elements import Schematic
from .DiagramTranslator import SchematicDiagramParser, circuit_translator
from ..Circuit.solution import ComplexSolution
import schemdraw
import matplotlib as mpl
from typing import Callable
import numpy as np

def linear_colormap(start_color: tuple[float, float, float], end_color: tuple[float, float, float], num_samples: int) -> Callable[[float], tuple[float, float, float]]:
    return mpl.colors.LinearSegmentedColormap.from_list('light_bulb', [start_color, end_color])

def light_lamps(schematic: Schematic, on_threshold_percentage: float = 0.05, breakthrough_threshold_percentage: float = 1.2) -> None:
    def light_color(brightness: float) -> tuple[float, float, float]:
        colormap: Callable[[float], tuple[float, float, float]] = linear_colormap((1.0, 0.62, 0.24), (0.97, 1.0, 0.52), 256)
        if brightness <= on_threshold_percentage:
            return (1, 1, 1)
        if brightness >= breakthrough_threshold_percentage:
            return (0.2, 0.2, 0.2)
        return colormap(brightness)
    diagram_parser = SchematicDiagramParser(schematic)
    solution = ComplexSolution(circuit=circuit_translator(schematic))
    lamps = [component for component in solution.circuit.components if component.type == 'lamp']
    brightness_percentages = [solution.get_power(lamp.id).real/float(lamp.value['P']) for lamp in lamps]
    for lamp, brght_pct in zip(lamps, brightness_percentages):
        schemdraw_lamp = diagram_parser.get_element(lamp.id)
        schemdraw_lamp.segments.append(schemdraw.SegmentCircle((0.5, 0), 0.5, color=light_color(brght_pct), fill=True, zorder=-100))


from ..Network import network as ntw
from ..Network import elements as ntw_elm

from . import Elements as elm
from .SchemdrawTranslatorTypes import ElementTranslatorMap


def linear_current_source_translator(element: elm.RealCurrentSource, nodes: tuple[str, str]) -> ntw.Branch:
    return ntw.Branch(nodes[0], nodes[1], ntw_elm.linear_current_source(I=element.I, Y=1/element.R, name=element.name))

def resistor_translator(element: elm.Resistor, nodes: tuple[str, str]) -> ntw.Branch:
    return ntw.Branch(nodes[0], nodes[1], ntw_elm.resistor(R=element.R, name=element.name))

def impedance_translator(element: elm.Resistor, nodes: tuple[str, str]) -> ntw.Branch:
    return ntw.Branch(nodes[0], nodes[1], ntw_elm.impedance(Z=element.Z, name=element.name))

def current_source_translator(element: elm.CurrentSource, nodes: tuple[str, str]) -> ntw.Branch:
    return ntw.Branch(nodes[0], nodes[1], ntw_elm.current_source(I=element.I, name=element.name))

def linear_voltage_source_translator(element: elm.RealVoltageSource, nodes: tuple[str, str]) -> ntw.Branch:
    return ntw.Branch(nodes[0], nodes[1], ntw_elm.linear_voltage_source(V=-element.V, Z=element.R, name=element.name))

def voltage_source_translator(element: elm.VoltageSource, nodes: tuple[str, str]) -> ntw.Branch:
    return ntw.Branch(nodes[0], nodes[1], ntw_elm.voltage_source(V=-element.V, name=element.name))

def none_translator(*_) -> None:
    return None

network_translator_map : ElementTranslatorMap = {
    elm.Resistor : resistor_translator,
    elm.Impedance : impedance_translator,
    elm.CurrentSource: current_source_translator,
    elm.VoltageSource: voltage_source_translator,
    elm.RealCurrentSource: linear_current_source_translator,
    elm.RealVoltageSource: linear_voltage_source_translator,
    elm.Line: none_translator,
    elm.Node: none_translator,
    elm.LabelNode: none_translator,
    elm.Ground: none_translator,
}



from .components import Component, is_active
from .transformers import transformers
from ..Network.network import Network
import numpy as np
from dataclasses import dataclass, field

class MultipleGroundNodes(Exception): pass
class AmbiguousComponentID(Exception): pass

@dataclass
class Circuit:
    components : list[Component]
    ground_node : str = field(init=False, default='')

    def __post_init__(self) -> None:
        if len(self.components) == 0:
            self.ground_node = ''
            return
        ground_nodes = [component.nodes[0] for component in self.components if component.type == 'ground']
        if len(ground_nodes) > 1:
            raise MultipleGroundNodes(f'Component list contains multiple ground nodes: {str(ground_nodes)}')
        if len(ground_nodes) == 0:
            self.ground_node = self.components[0].nodes[0]
        else:
            self.ground_node = ground_nodes[0]
        if len(set([component.id for component in self.components])) != len(self.components):
            raise AmbiguousComponentID(f'Component list contains multiple components with the same ID.')

    def __getitem__(self, key: str) -> Component:
        index = [component.id for component in self.components].index(key)
        return self.components[index]

def w(f: float) -> float:
    return 2*np.pi*f

def transform_circuit(circuit: Circuit, w: float, w_resolution: float = 1e-3) -> Network:
    return Network(
        branches=[transformers[component.type](component, w, w_resolution) for component in circuit.components if component.type in transformers.keys()],
        node_zero_label=circuit.ground_node
    )

def transform(circuit: Circuit, w: list[float] = [0], w_resolution: float = 1e-3) -> list[Network]:
    return [transform_circuit(circuit, w_, w_resolution) for w_ in w]

def frequency_components(circuit: Circuit, w_max: float) -> list[float]:
    def frequencies(component: Component) -> list[float]:
        try:
            w = float(component.value['w'])
        except KeyError:
            return []
        if component.type == 'periodic_voltage_source' or component.type == 'periodic_current_source':
            n_max = np.floor(w_max/w)
            return [w*n for n in np.arange(n_max+1)]
        return [w]
    return sorted(list(set([w for c in circuit.components for w in frequencies(c)])))


from typing import Any, Callable
from .network import Network, Branch
from . import elements as elm
import numpy as np

class FileFormatError(Exception):
    ...

def to_complex(z: dict[str, float], degree: bool = False) -> complex:
    try:
        return complex(z['real'], z['imag'])
    except (KeyError, TypeError):
        ...
    try:
        if degree:
            z['phase'] *= np.pi/180
        return z['abs']*complex(np.cos(z['phase']), np.sin(z['phase']))
    except (KeyError, TypeError):
        raise FileFormatError

def translate_to_complex(keys : list[str], **kwargs):
    for value in keys:
        kwargs.update({value : to_complex(kwargs[value])})
    return kwargs

network_branch_translators : dict[str, Callable[..., elm.NortenTheveninElement]] = {
    "resistor" : elm.resistor,
    "conductor" : elm.conductor,
    "impedance" : lambda **kwargs: elm.impedance(Z=to_complex(kwargs.pop('Z')), **kwargs),
    "admittance" : lambda **kwargs: elm.admittance(Y=to_complex(kwargs['Y']), **kwargs),
    "linear_current_source" : lambda **kwargs: elm.current_source(**translate_to_complex(keys=['I', 'Y'], **kwargs)),
    "current_source" : lambda **kwargs: elm.current_source(I=to_complex(kwargs.pop('I')), **kwargs),
    "real_current_source" : elm.current_source,
    "linear_voltage_source" : lambda **kwargs: elm.voltage_source(V=to_complex(kwargs.pop('V')), Z=to_complex(kwargs.pop('Z')), **kwargs),
    "voltage_source" : lambda **kwargs: elm.voltage_source(V=to_complex(kwargs.pop('V')), **kwargs),
    "real_voltage_source" : elm.voltage_source,
    "short_circuit" : elm.short_circuit,
    "open_circuit" : elm.open_circuit,
}

def load_network(network_dict: list[dict[str, Any]]) -> Network:
    def entry_to_branch(entry: dict[str, Any]) -> Branch:
        n1 = entry.pop('N1')
        n2 = entry.pop('N2')
        entry['name'] = entry.pop('id')
        element_factory = network_branch_translators[entry.pop('type')]
        element = element_factory(**entry)
        return Branch(n1, n2, element)
    try: 
        return Network([entry_to_branch(entry) for entry in network_dict])
    except KeyError:
        raise FileExistsError

def load_network_from_json(filename: str) -> Network:
    import json
    with open(filename, 'r') as json_file:
        network_dict = json.load(json_file)
    return load_network(network_dict)



import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from matplotlib.axes import Axes
import numpy as np
from functools import partial
from CircuitCalculator.Utils import scientific_float
from CircuitCalculator.Network.EquivalentSources import NortenEquivalentSource, TheveninEquivalentSource

def plot_equivalent_source_diagram(x0: float, y0: float, xlabel: str, ylabel: str, m2: float = 0, title: str = '', x0label: str = '', y0label: str = '') -> tuple[Figure, Axes]:
    fig, ax = plt.subplots()
    ax.plot([0, x0], [y0, 0], linewidth=2)
    ax.grid(visible=True)
    ax.set_xlim(left=0, right=ax.get_xticks()[-1])
    ax.set_ylim(bottom=0, top=ax.get_yticks()[-1])
    ax.set_xlabel(f'${xlabel} \\rightarrow$')
    ax.set_ylabel(f'${ylabel} \\rightarrow$')
    ax.set_title(title)
    
    xticks = ax.get_xticks()
    xticks = np.sort(np.append(xticks, x0))
    ax.set_xticks(xticks)
    
    yticks = ax.get_yticks()
    yticks = np.sort(np.append(yticks, y0))
    ax.set_yticks(yticks)

    axes_text = partial(ax.text, rotation_mode='anchor', ha='left', va='center', backgroundcolor='white')
    axes_text(
        x=x0,
        s=x0label,
        y=1/2*np.diff(ax.get_yticks()[:2])[0],
        rotation='vertical'
    )
    axes_text(
        x=1/2*np.diff(ax.get_xticks()[:2])[0],
        y=y0,
        s=y0label
    )

    if m2 > 0:
        xmax = ax.get_xlim()[-1]
        ax.plot([0, xmax], [0, m2*xmax], color='red')
        x_load = y0/(y0/x0+m2)
        y_load = m2*x_load
        ax.plot(x_load, y_load, marker='o', color='red')

    fig.show()
    return fig, ax

def plot_thevenin_source(source: TheveninEquivalentSource, xlabel: str = 'I', ylabel: str = 'U', title: str = 'Thevenin Equivalent Source', R_load: float = 0) -> tuple[Figure, Axes]:
    UL = np.real(source.U)
    Ri = np.real(source.Z)
    IK = UL/Ri
    sf = partial(scientific_float, exp_prefixes={3: 'k', -3: 'm', -6: 'u'})
    fig, ax = plot_equivalent_source_diagram(
        x0=IK,
        y0=UL,
        xlabel=xlabel,
        ylabel=ylabel,
        m2=R_load,
        title=title,
        x0label='$I_\\mathrm{K}='+sf(IK)+'\\mathrm{A}$',
        y0label='$U_\\mathrm{L}='+sf(UL)+'\\mathrm{V}$'
    )
    fig.show()
    return fig, ax

def plot_norten_source(source: NortenEquivalentSource, xlabel: str = 'U', ylabel: str = 'I', title: str = 'Norten Equivalent Source', R_load: float = 0) -> tuple[Figure, Axes]:
    IK = np.real(source.I)
    Ri = 1/np.real(source.Y)
    UL = Ri*IK
    sf = partial(scientific_float, exp_prefixes={3: 'k', -3: 'm', -6: 'u'})
    fig, ax = plot_equivalent_source_diagram(
        x0=UL,
        y0=IK,
        xlabel=xlabel,
        ylabel=ylabel,
        m2=1/R_load,
        title=title,
        x0label='$U_\\mathrm{L}='+sf(UL)+'\\mathrm{V}$',
        y0label='$I_\\mathrm{K}='+sf(IK)+'\\mathrm{A}$'
    )
    fig.show()
    return fig, ax


from .node_analysis import nodal_analysis_coefficient_matrix, nodal_analysis_constants_vector, open_circuit_impedance
from ..elements import is_ideal_current_source, is_current_source
from ..network import Network
from ..solution import NetworkSolution
from .solution import NodalAnalysisSolution
import numpy as np
from dataclasses import dataclass

@dataclass
class NodalAnalysisBiasPointSolution(NodalAnalysisSolution):
    def __post_init__(self) -> None:
        A = nodal_analysis_coefficient_matrix(self.network, node_mapper=self.node_mapper)
        b = nodal_analysis_constants_vector(self.network, node_mapper=self.node_mapper)
        try:
            self._solution_vector = np.linalg.solve(A, b)
        except np.linalg.LinAlgError:
            self._solution_vector = np.zeros(np.size(b))
        if np.any(np.isnan(self._solution_vector)):
            self._solution_vector = np.zeros(np.size(b))

    @property
    def _potentials(self) -> np.ndarray:
        return self._solution_vector[:self._node_mapping.N]

    @property
    def _voltage_source_currents(self) -> np.ndarray:
        return self._solution_vector[-self._voltage_source_mapping.N:]

    def get_potential(self, node_id: str) -> complex:
        if node_id == self.network.node_zero_label:
            return 0
        return self._potentials[self._node_mapping[node_id]]

    def get_current(self, branch_id: str) -> complex:
        if branch_id in self._voltage_source_mapping.keys:
            return self._voltage_source_currents[self._voltage_source_mapping[branch_id]]
        if is_ideal_current_source(self.network[branch_id].element):
            return self.network[branch_id].element.I
        if is_current_source(self.network[branch_id].element):
            return - (self.network[branch_id].element.I + self.get_voltage(branch_id)/self.network[branch_id].element.Z)
        branch = self.network[branch_id]
        return self.get_voltage(branch_id)/branch.element.Z

def open_circuit_voltage(network: Network, node1: str, node2: str) -> complex:
    solution = NodalAnalysisBiasPointSolution(network)
    if node1 == node2:
        return 0
    phi1 = solution.get_potential(node_id=node1)
    phi2 = solution.get_potential(node_id=node2)
    return phi1-phi2

def short_circuit_current(network: Network, node1: str, node2: str) -> complex:
    Z = open_circuit_impedance(network, node1, node2)
    V = open_circuit_voltage(network, node1, node2)
    return V/Z

def nodal_analysis_bias_point_solver(network: Network) -> NetworkSolution:
    return NodalAnalysisBiasPointSolution(network)


from typing import Protocol
from dataclasses import dataclass
import numpy as np

class NortenTheveninElement(Protocol):
    @property
    def name(self) -> str:
        """Name of element"""
        ...
    @property
    def type(self) -> str:
        """Specific element type"""
        ...
    @property
    def Z(self) -> complex:
        """Impedance value of element"""
        ...
    @property
    def Y(self) -> complex:
        """Admittance value of element"""
        ...
    @property
    def V(self) -> complex:
        """Voltage value of element"""
        ...
    @property
    def I(self) -> complex:
        """Current value of element"""
        ...

@dataclass(frozen=True)
class NortenElement:
    name : str
    type : str
    Z : complex
    V : complex

    @property
    def Y(self) -> complex:
        try:
            return 1/self.Z
        except ZeroDivisionError:
            return np.inf

    @property
    def I(self) -> complex:
        try:
            return complex(self.V)/self.Z
        except ZeroDivisionError:
            return np.nan

@dataclass(frozen=True)
class TheveninElement:
    name : str
    type : str
    Y : complex
    I : complex

    @property
    def Z(self) -> complex:
        try:
            return 1/self.Y
        except ZeroDivisionError:
            return np.inf

    @property
    def V(self) -> complex:
        try:
            return complex(self.I)/self.Y
        except ZeroDivisionError:
            return np.nan

def impedance(name : str, Z : complex) -> NortenTheveninElement:
    return NortenElement(Z=Z, V=0, name=name, type='impedance')

def admittance(name : str, Y : complex) -> NortenTheveninElement:
    return TheveninElement(Y=Y, I=0, name=name, type='admittance')

def resistor(name : str, R : float) -> NortenTheveninElement:
    return NortenElement(Z=R, V=0, name=name, type='resistor')

def conductor(name : str, G : float) -> NortenTheveninElement:
    return TheveninElement(Y=G, I=0, name=name, type='conductor')

def load(name : str, P : float, V_ref: float = -1, I_ref: float = -1, Q : float = 0) -> NortenTheveninElement:
    if V_ref < 0 and I_ref < 0:
        raise AttributeError('A reference voltage or reference current has to be defined for a load.')
    if V_ref > 0 and I_ref > 0:
        raise AttributeError('You can only define a reference voltage or a reference current for a load.')
    if V_ref == 0 and I_ref < 0:
        raise ValueError('Reference voltage must be greater than zero.')
    if I_ref <= 0 and V_ref < 0:
        raise ValueError('Reference current must be greater than zero.')
    if V_ref > 0:
        return TheveninElement(Y=complex(P, Q)/V_ref**2, I=0, name=name, type='load')
    return NortenElement(Z=complex(P, Q)/I_ref**2, V=0, name=name, type='load')

def voltage_source(name : str, V : complex, Z : complex = 0) -> NortenTheveninElement:
    return NortenElement(V=V, Z=Z, name=name, type='voltage_source')

def current_source(name : str, I : complex, Y : complex = 0) -> NortenTheveninElement:
    return TheveninElement(I=I, Y=Y, name=name, type='current_source')

def open_circuit(name : str) -> NortenTheveninElement:
    return TheveninElement(I=0, Y=0, name=name, type='open_circuit')

def short_circuit(name : str) -> NortenTheveninElement:
    return NortenElement(V=0, Z=0, name=name, type='short_circuit')

def impedance_value(R : float = 0.0, X : float = 0.0, absZ : float = -1.0, phi : float = 0.0, degree : bool = False) -> complex:
    if degree:
        phi *= np.pi/180
    if absZ > 0:
        return complex(absZ*np.cos(phi), absZ*np.sin(phi))
    return complex(R, X)

def admittance_value(G : float = 0.0, B : float = 0.0, absY : float = -1.0, phi : float = 0.0, degree : bool = False) -> complex:
    if degree:
        phi *= np.pi/180
    if absY > 0:
        return complex(absY*np.cos(phi), absY*np.sin(phi))
    return complex(G, B)

def complex_value(X : float, phi : float = 0.0, rms: bool = False, deg: bool = False):
    if rms:
        X *= np.sqrt(2)
    if deg:
        phi = np.radians(phi)
    if not np.isfinite(X):
        return complex(np.inf, 0)
    return X*complex(np.cos(phi), np.sin(phi))

def is_voltage_source(element: NortenTheveninElement) -> bool:
    return np.abs(element.V) > 0

def is_current_source(element: NortenTheveninElement) -> bool:
    return np.abs(element.I) > 0

def is_ideal_voltage_source(element: NortenTheveninElement) -> bool:
    return np.abs(element.V) >= 0 and element.Z==0

def is_ideal_current_source(element: NortenTheveninElement) -> bool:
    return np.abs(element.I) >= 0 and element.Y==0

def is_active(element: NortenTheveninElement) -> bool:
    return is_voltage_source(element) or is_current_source(element)

def is_short_circuit(element: NortenTheveninElement) -> bool:
    return element.V == 0 and element.Z == 0

def is_open_circuit(element: NortenTheveninElement) -> bool:
    return element.I == 0 and element.Y == 0




from typing import TypeVar, Callable, Any, Type, Union
import schemdraw.elements
from ..Circuit.components import Component

SchemdrawElement = TypeVar('SchemdrawElement', bound=schemdraw.elements.Element)
SchemdrawElementTranslator = Callable[[SchemdrawElement, tuple[str, ...]], Component | None] | Callable[[SchemdrawElement, tuple[str, ...]], Component | None]

ElementTranslatorMap = dict[Type[schemdraw.elements.Element], SchemdrawElementTranslator]


from dataclasses import dataclass
import schemdraw.elements
from ..Network.network import Network
from ..Circuit.circuit import Circuit, Component

from . import Elements as elm
from .CircuitComponentTranslators import circuit_translator_map
from .NetworkBranchTranslators import network_translator_map
from .SchemdrawTranslatorTypes import ElementTranslatorMap

from .DiagramParser import SchematicDiagramParser

class UnknownTranslator(Exception): pass

@dataclass
class DiagramTranslator:
    diagram_parser : SchematicDiagramParser
    translator_map : ElementTranslatorMap

    def __call__(self, element: schemdraw.elements.Element) -> Component | None:
        try:
            return self.translator_map[type(element)](element, tuple(map(self.diagram_parser._get_node_index, elm.get_nodes(element))))
        except KeyError:
            raise UnknownTranslator(f"Element '{type(element).__name__}' cannot be translated.")

def _remove_none(l: list) -> list:
    return [e for e in l if e is not None]

def circuit_translator(schematic: elm.Schematic) -> Circuit:
    parser = SchematicDiagramParser(schematic)
    translator = DiagramTranslator(parser, circuit_translator_map)
    return Circuit(_remove_none([translator(e) for e in parser.all_elements]))

def network_translator(schematic: elm.Schematic) -> Network:
    parser = SchematicDiagramParser(schematic)
    translator = DiagramTranslator(parser, network_translator_map)
    return Network(_remove_none([translator(e) for e in parser.all_elements]), parser.ground_label)


import schemdraw, schemdraw.elements
from . import Display as dsp
from typing import Type, Optional

def resistor(element: Type[schemdraw.elements.Element2Term]) -> Type[schemdraw.elements.Element2Term]:
    class extended_resistor(element):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.anchors['value_label'] = (0.5, 0.3)
            self.anchors['v_label'] = (0.5, -1.1)
            self.anchors['s_label'] = (0.5, 0.9)

        def down(self, length: Optional[float] = None) -> schemdraw.elements.Element:
            self.anchors['value_label'] = (0.5, -0.6)
            self.anchors['v_label'] = (0.5, 0.3)
            self.anchors['s_label'] = (0.5, -1.2)
            return super().down(length=length)

        def up(self, length: Optional[float] = None) -> schemdraw.elements.Element:
            self.anchors['value_label'] = (0.5, 0.4)
            self.anchors['v_label'] = (0.5, -1.1)
            self.anchors['s_label'] = (0.5, -1.2)
            return super().up(length=length)

        def left(self, length: Optional[float] = None) -> schemdraw.elements.Element:
            self.anchors['value_label'] = (0.5, -0.9)
            self.anchors['v_label'] = (0.5, 0.3)
            self.anchors['s_label'] = (0.5, -1.2)
            return super().left(length=length)

        def _place_label(self, label: schemdraw.elements.elements.Label, theta: float = 0):
            delta = self.end-self.start
            if abs(delta[1]) > abs(delta[0]): # portrait placing
                if delta[1] < 0:
                    theta = 90
            super()._place_label(label=label, theta=theta)
    
    return extended_resistor

def source(element: Type[schemdraw.elements.Element2Term]) -> Type[schemdraw.elements.Element2Term]:
    class extended_source(element):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.anchors['value_label'] = (0.5, 1.1)
            self.anchors['v_label'] = (0.5, -1.1)
            self.anchors['i_label'] = (1.2, 0.3)
            self.anchors['s_label'] = (0.5, 1.5)

        def down(self, length: Optional[float] = None) -> schemdraw.elements.Element:
            self.anchors['s_label'] = (0.5, -0.7)
            return super().down(length=length)
    return extended_source

def capacitor(element: Type[schemdraw.elements.Element2Term]) -> Type[schemdraw.elements.Element2Term]:
    class extended_capacitor(element):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.anchors['value_label'] = (0.0, 0.3)
            self.anchors['v_label'] = (0.0, -1.0)
            self.anchors['s_label'] = (0.0, 0.9)

        def down(self, length: Optional[float] = None) -> schemdraw.elements.Element:
            self.anchors['value_label'] = (0.0, -0.6)
            self.anchors['s_label'] = (0.0, -1.1)
            self.anchors['v_label'] = (0.0, 0.3)
            return super().down(length=length)

        def left(self, length: Optional[float] = None) -> schemdraw.elements.Element:
            self.anchors['value_label'] = (0.0, -0.6)
            self.anchors['s_label'] = (0.0, -1.1)
            self.anchors['v_label'] = (0.0, 0.3)
            return super().left(length=length)

        def _place_label(self, label: schemdraw.elements.elements.Label, theta: float = 0):
            delta = self.end-self.start
            if abs(delta[1]) > abs(delta[0]): # portrait placing
                if delta[1] < 0:
                    theta = 90
            super()._place_label(label, theta)
    return extended_capacitor

def inductor(element: Type[schemdraw.elements.Element2Term]) -> Type[schemdraw.elements.Element2Term]:
    class extended_inductor(element):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.anchors['value_label'] = (0.5, 0.3)
            self.anchors['v_label'] = (0.5, -0.8)
            self.anchors['s_label'] = (0.5, 0.9)

        def down(self, length: Optional[float] = None) -> schemdraw.elements.Element:
            self.anchors['value_label'] = (0.5, -0.6)
            self.anchors['v_label'] = (0.5, 0.3)
            self.anchors['s_label'] = (0.5, -1.2)
            return super().down(length=length)

        def left(self, length: Optional[float] = None) -> schemdraw.elements.Element:
            self.anchors['value_label'] = (0.5, -0.6)
            self.anchors['v_label'] = (0.5, 0.2)
            self.anchors['s_label'] = (0.5, -1.2)
            return super().left(length=length)

        def _place_label(self, label: schemdraw.elements.elements.Label, theta: float = 0):
            delta = self.end-self.start
            if abs(delta[1]) > abs(delta[0]): # portrait placing
                if delta[1] < 0:
                    theta = 90
            super()._place_label(label, theta)
    return extended_inductor

def linear_current_source(element: Type[schemdraw.elements.Element2Term]) -> Type[schemdraw.elements.Element2Term]:
    class extended_linear_source(element):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.anchors['value_label'] = (0.5, -1.2)
            self.anchors['start'] = (0, 0)
            self.anchors['center'] = (0, 1.5)
            self.anchors['end'] = (3, 0)

        def up(self, length: Optional[float] = None) -> schemdraw.elements.Element:
            self.anchors['value_label'] = (0.5, -0.6)
            self.anchors['v_label'] = (1.5, -2.5)
            self.anchors['s_label'] = (0.5, -1.2)
            return super().up(length=length)

    return extended_linear_source

def linear_voltage_source(element: Type[schemdraw.elements.Element2Term]) -> Type[schemdraw.elements.Element2Term]:
    class extended_linear_source(element):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.anchors['v_label'] = (2, -1.5)
            self.anchors['start'] = (-5, 0)
            self.anchors['center'] = (0, 1.5)
            self.anchors['end'] = (0, 0)

        def up(self, length: Optional[float] = None) -> schemdraw.elements.Element:
            self.anchors['v_label'] = (-1.5, -2.5)
            return super().up(length=length)

    return extended_linear_source

def lamp(element: Type[schemdraw.elements.Element2Term]) -> Type[schemdraw.elements.Element2Term]:
    class extended_lamp(element):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.anchors['value_label'] = (0.5, 0.8)
            self.anchors['v_label'] = (0.5, -1.1)
            self.anchors['i_label'] = (1.2, 0.3)
            self.anchors['s_label'] = (0.5, 1.5)

        def down(self, length: Optional[float] = None) -> schemdraw.elements.Element:
            self.anchors['s_label'] = (0.5, -0.7)
            return super().down(length=length)

    return extended_lamp

def voltage_arrow(start: tuple[float, float] = (1.5, 0.7), end: tuple[float, float] = (-0.5, 0.7), arrowwidth: float = 0.3, arrowlength: float = 0.4, color=dsp.blue) -> schemdraw.Segment:
    return schemdraw.Segment((start, end), arrow='->', arrowwidth=arrowwidth, arrowlength=arrowlength, color=color)

def current_arrow(start: tuple[float, float] = (1.2, 0.3), end: tuple[float, float] = (1.8, 0.3), arrowwidth: float = 0.3, arrowlength: float = 0.4, color=dsp.red) -> schemdraw.Segment:
    return schemdraw.Segment((start, end), arrow='->', arrowwidth=arrowwidth, arrowlength=arrowlength, color=color)



from .network import Network
from typing import Callable, Protocol, Any

class NetworkSolution(Protocol):
    def get_voltage(self, branch_id: str) -> Any: ...

    def get_current(self, branch_id: str) -> Any: ...

    def get_potential(self, node_id: str) -> Any: ...

    def get_power(self, branch_id: str) -> Any: ...

NetworkSolver = Callable[[Network], NetworkSolution]


from schemdraw.elements.twoterm import ResistorIEC
from schemdraw.elements.sources import Source
from schemdraw.transform import Transform
from schemdraw.segments import Segment

class InductorDIN(ResistorIEC):
    def __init__(self, *d, **kwargs):
        super().__init__(*d, **kwargs)
        self.segments = [self.segments[0].xform(Transform(theta=0, globalshift=(0,0)), fill='black')]

class SourceUDIN(Source):
    def __init__(self, *d, **kwargs):
        super().__init__(*d, **kwargs)
        self.segments.append(Segment([(0, 0), (1, 0)]))

class SourceIDIN(Source):
    def __init__(self, *d, **kwargs):
        super().__init__(*d, **kwargs)
        self.segments.append(Segment([(0.5, -0.5), (0.5, 0.5)]))

class SourceIDIN2(SourceIDIN):
    def __init__(self, *d, **kwargs):
        super().__init__(*d, **kwargs)
        self.segments.append(Segment((1.2, 0.2), (1.7, 0.2), arrow='>'))



from ..Network.solution import NetworkSolver, NetworkSolution
from ..Network.NodalAnalysis.bias_point_analysis import nodal_analysis_bias_point_solver
from ..Circuit.solution import ComplexSolution, DCSolution

from . import Elements as elm
from . import Display as dsp
from .DiagramTranslator import SchematicDiagramParser, circuit_translator

from dataclasses import dataclass
from typing import Callable
from functools import partial

class UnknownElement(Exception): pass
class MultipleGroundNodes(Exception): pass
class UnknownTranslator(Exception): pass

@dataclass
class SchematicDiagramSolution:
    diagram_parser: SchematicDiagramParser
    solution: ComplexSolution | NetworkSolution
    voltage_display: Callable[[complex], str]
    current_display: Callable[[complex], str]
    power_display: Callable[[complex], str]

    def draw_voltage(self, name: str, reverse: bool = False) -> elm.VoltageLabel:
        element = self.diagram_parser.get_element(name)
        V_branch = self.solution.get_voltage(name)
        if reverse:
            V_branch *= -1
        return elm.VoltageLabel(element, vlabel=self.voltage_display(V_branch), reverse=reverse if not element.is_reverse else not reverse, color=dsp.blue)

    def draw_current(self, name: str, reverse: bool = False, end: bool = False) -> elm.CurrentLabel:
        element = self.diagram_parser.get_element(name)
        I_branch = self.solution.get_current(name)
        if reverse:
            I_branch *= -1
        if end:
            reverse = not reverse
        return elm.CurrentLabel(element, ilabel=self.current_display(I_branch), reverse=reverse if not element.is_reverse else not reverse, start=not end, color=dsp.red)

    def draw_power(self, name: str, reverse: bool = False) -> elm.PowerLabel:
        element = self.diagram_parser.get_element(name)
        P_branch = self.solution.get_power(name)
        return elm.PowerLabel(element, plabel=self.power_display(P_branch), color=dsp.green)

    def draw_potential(self, node: str, loc:str = '') -> elm.LabelNode:
        element = self.diagram_parser.get_element(node)
        node_potential = self.solution.get_potential(node_id=node)
        return elm.LabelNode(id_loc=loc, name=self.voltage_display(node_potential), at=element.absdrop[0], color=dsp.blue)

def time_domain_steady_state_solution(schematic: elm.Schematic, solver: NetworkSolver = nodal_analysis_bias_point_solver, w: float = 0, sin: bool = False, deg: bool = False, hertz: bool = False) -> SchematicDiagramSolution:
    digagram_parser = SchematicDiagramParser(schematic)
    solution = ComplexSolution(circuit=circuit_translator(schematic), solver=solver, w=w)
    return SchematicDiagramSolution(
        diagram_parser=digagram_parser,
        solution=solution,
        voltage_display=partial(dsp.print_sinosoidal, unit='V', w=w, sin=sin, deg=deg, hertz=hertz),
        current_display=partial(dsp.print_sinosoidal, unit='A', w=w, sin=sin, deg=deg, hertz=hertz),
        power_display=dsp.print_active_reactive_power
    )

def single_frequency_solution(schematic: elm.Schematic, solver: NetworkSolver = nodal_analysis_bias_point_solver, w: float = 0, precision: int = 3, polar: bool = False, deg: bool = False) -> SchematicDiagramSolution:
    diagram_parser = SchematicDiagramParser(schematic)
    solution = ComplexSolution(circuit=circuit_translator(schematic), solver=solver, w=w)
    return SchematicDiagramSolution(
        diagram_parser=diagram_parser,
        solution=solution,
        voltage_display=partial(dsp.print_complex, unit='V', precision=precision, polar=polar, deg=deg),
        current_display=partial(dsp.print_complex, unit='A', precision=precision, polar=polar, deg=deg),
        power_display=partial(dsp.print_complex, unit='VA', precision=precision, polar=polar, deg=deg)
    )

def complex_network_dc_solution(schematic: elm.Schematic, solver: NetworkSolver = nodal_analysis_bias_point_solver, precision: int = 3) -> SchematicDiagramSolution:
    diagram_parser = SchematicDiagramParser(schematic)
    solution = ComplexSolution(circuit=circuit_translator(schematic), solver=solver)
    return SchematicDiagramSolution(
        diagram_parser=diagram_parser,
        solution=solution,
        voltage_display=partial(dsp.print_complex, unit='V', precision=precision),
        current_display=partial(dsp.print_complex, unit='A', precision=precision),
        power_display=partial(dsp.print_complex, unit='VA', precision=precision)
    )

def real_network_dc_solution(schematic: elm.Schematic, solver: NetworkSolver = nodal_analysis_bias_point_solver, precision: int = 3) -> SchematicDiagramSolution:
    diagram_parser = SchematicDiagramParser(schematic)
    solution = DCSolution(circuit=circuit_translator(schematic), solver=solver)
    return SchematicDiagramSolution(
        diagram_parser=diagram_parser,
        solution=solution,
        voltage_display=partial(dsp.print_real, unit='V', precision=precision),
        current_display=partial(dsp.print_real, unit='A', precision=precision),
        power_display=lambda x: dsp.print_active_power(x.real, precision=precision)
    )


from .network import Network, Branch
from .elements import NortenTheveninElement, is_current_source, is_voltage_source, is_short_circuit, is_open_circuit, impedance, admittance, is_open_circuit, is_short_circuit

def switch_ground_node(network: Network, new_ground: str) -> Network:
    return Network(network.branches, new_ground)

def remove_element(network: Network, element: str) -> Network:
    branches = list(network.branches)
    branches.remove(network[element])
    return Network(branches, node_zero_label=network.node_zero_label)

def remove_open_circuit_elements(network: Network) -> Network:
    return Network([b for b in network.branches if not is_open_circuit(b.element)], node_zero_label=network.node_zero_label)

def remove_short_circuit_elements(network: Network, keep: list[NortenTheveninElement] = []) -> Network:
    branches = network.branches
    short_circuits = [b for b in network.branches if is_short_circuit(b.element) and b.element not in keep]
    short_circuit_nodes = [(vs.node1, vs.node2) if not network.is_zero_node(vs.node1) else (vs.node2, vs.node1) for vs in short_circuits]
    for an, rn in short_circuit_nodes:
        branches = [Branch(rn, b.node2, b.element) if b.node1 == an else b for b in branches]
        branches = [Branch(b.node1, rn, b.element) if b.node2 == an else b for b in branches]
        branches = [b for b in branches if b.node1 != b.node2]
    return Network(branches, node_zero_label=network.node_zero_label)

def short_circuitify_voltage_sources(network: Network, keep: list[NortenTheveninElement] = []) -> Network:
    def zero_in_voltage(branch: Branch) -> Branch:
        return Branch(branch.node1, branch.node2, impedance(branch.element.name, branch.element.Z))
    def is_intended_voltage_source(branch: Branch) -> bool:
        if branch.element in keep:
            return False
        if is_voltage_source(branch.element):
            return True
        return False
    return Network(
        branches=[zero_in_voltage(b) if is_intended_voltage_source(b) else b for b in network.branches],
        node_zero_label=network.node_zero_label
    )

def open_circuitify_current_sources(network: Network, keep: list[NortenTheveninElement] = []) -> Network:
    def zero_in_current(branch: Branch) -> Branch:
        return Branch(branch.node1, branch.node2, admittance(branch.element.name, branch.element.Y))
    def is_intended_current_source(branch: Branch) -> bool:
        if branch.element in keep:
            return False
        if is_current_source(branch.element):
            return True
        return False
    return Network(
        branches=[zero_in_current(b) if is_intended_current_source(b) else b for b in network.branches],
        node_zero_label=network.node_zero_label
    )

def remove_ideal_current_sources(network: Network, keep: list[NortenTheveninElement] = []) -> Network:
    return remove_open_circuit_elements(open_circuitify_current_sources(network, keep=keep))

def remove_ideal_voltage_sources(network: Network, keep: list[NortenTheveninElement] = []) -> Network:
    return remove_short_circuit_elements(short_circuitify_voltage_sources(network, keep=keep), keep=keep)

def passive_network(network: Network, keep: list[NortenTheveninElement] = []) -> Network:
    return remove_ideal_voltage_sources(remove_ideal_current_sources(network, keep=keep), keep=keep)


import schemdraw.elements, schemdraw.util
from dataclasses import dataclass
from . import Elements as elm

class UnknownElement(Exception): pass
class MultipleGroundNodes(Exception): pass

@dataclass(frozen=True)
class SchematicDiagramParser:
    drawing: elm.Schematic

    @property
    def all_elements(self) -> list[schemdraw.elements.Element]:
        return self.drawing.elements

    @property
    def circuit_elements(self) -> list[schemdraw.elements.Element]:
        def has_name_property(e: schemdraw.elements.Element):
            try:
                _ = e.name
            except AttributeError:
                return False
            return True
        return [e for e in self.all_elements if has_name_property(e)]

    @property
    def line_elements(self) -> list[elm.Line]:
        return [e for e in self.all_elements if type(e) is elm.Line]
    
    @property
    def node_elements(self) -> list[elm.Node]:
        return [e for e in self.all_elements if isinstance(e, elm.Node)]

    @property
    def all_nodes(self) -> set[schemdraw.util.Point]:
        nodes = {elm.round_node(e.absanchors['start']) for e in self.circuit_elements}
        nodes = nodes.union({elm.round_node(e.absanchors['end']) for e in self.circuit_elements})
        nodes = nodes.union({elm.round_node(e.absanchors['start']) for e in self.line_elements})
        nodes = nodes.union({elm.round_node(e.absanchors['end']) for e in self.line_elements})
        return nodes

    @property
    def unique_nodes(self) -> set[schemdraw.util.Point]:
        nodes = self.all_nodes
        for node in self.all_nodes:
            if node in nodes:
                for n in self._get_equal_electrical_potential_nodes(node).intersection(nodes):
                    nodes.remove(n)
                nodes.add(node)
        return nodes

    @property
    def node_label_mapping(self) -> dict[schemdraw.util.Point, str]:
        node_labels = {self.unique_node_mapping[elm.get_nodes(e)[0]] : e.node_id for e in self.node_elements}
        node_index = len(node_labels)+1
        unlabeled_nodes = [p for p in self.unique_nodes if p not in node_labels.keys()]
        for p in unlabeled_nodes:
            while str(node_index) in node_labels.values():
                node_index += 1
            node_labels.update({p : str(node_index)})
        return node_labels

    @property
    def unique_node_mapping(self) -> dict[schemdraw.util.Point, schemdraw.util.Point]:
        node_mapping = {}
        for n in self.all_nodes:
            identical_nodes = self._get_equal_electrical_potential_nodes(n)
            identical_nodes.remove(n)
            current_unique_node = self.unique_nodes.intersection(identical_nodes)
            if len(current_unique_node) > 0:
                node_mapping.update({n: current_unique_node.pop()})
            else:
                node_mapping.update({n: n})
        return node_mapping

    @property
    def ground(self) -> schemdraw.util.Point:
        ground_nodes = [n for n in self.node_elements if isinstance(n, elm.Ground)]
        if len(ground_nodes) > 1:
            raise MultipleGroundNodes
        if len(ground_nodes) == 0:
            return list(self.unique_nodes)[0]
        else:
            return elm.get_nodes(ground_nodes[0])[0]

    @property
    def ground_label(self) -> str:
        return self._get_node_index(self.ground)

    def _get_equal_electrical_potential_nodes(self, node: schemdraw.util.Point) -> set[schemdraw.util.Point]:
        equal_electrical_potential_nodes = set([node])
        old_length = 0
        while len(equal_electrical_potential_nodes) > old_length:
            old_length = len(equal_electrical_potential_nodes)
            for line in self.line_elements:
                n1, n2 = elm.get_nodes(line)
                if n1 in equal_electrical_potential_nodes:
                    equal_electrical_potential_nodes.add(n2)
                elif n2 in equal_electrical_potential_nodes:
                    equal_electrical_potential_nodes.add(n1)
        return equal_electrical_potential_nodes

    def _get_node_index(self, node: schemdraw.util.Point) -> str:
        return self.node_label_mapping[self.unique_node_mapping[node]]

    def get_element(self, name: str) -> schemdraw.elements.Element:
        elements = [e for e in self.circuit_elements if e.name == name]
        if len(elements) == 0:
            raise UnknownElement(f'Element {name} not known')
        else:
            return elements[0]


from .network import Network
from .NodalAnalysis.node_analysis import open_circuit_impedance, open_circuit_voltage

class TheveninEquivalentSource:
    def __init__(self, network: Network, node1: str, node2: str) -> None:
        self.U = open_circuit_voltage(network, node1, node2)
        self.Z = open_circuit_impedance(network, node1, node2)

class NortenEquivalentSource:
    def __init__(self, network: Network, node1: str, node2: str) -> None:
        thevenin = TheveninEquivalentSource(network, node1, node2)
        self.I = thevenin.U/thevenin.Z
        self.Y = 1/thevenin.Z
        